 /*
 * extension.js
 *
 * Copyright 2014, 2015 the Widget Blocker extension copyright holders.
 * See the COPYRIGHT file at the top-level directory of this distribution for a
 * list of copyright holders and how to contact them.
 */

/*
 * This file is part of the Widget Blocker extension for GNOME Shell.
 *
 * Widget Blocker is free software: you are free to use, modify and/or
 * redistribute it for any purpose according to the terms of the LICENSE file
 * found in the top-level directory of this distribution.
 *
 * This program is distributed in the hope that it will be useful, WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the aforementioned LICENSE file for more details.
 */

const Main = imports.ui.main;
const Lang = imports.lang;
const LookingGlass = imports.ui.lookingGlass;
const Shell = imports.gi.Shell;
const Mainloop = imports.mainloop;
const Meta = imports.gi.Meta;
const ExtensionUtils = imports.misc.extensionUtils;
const Me = ExtensionUtils.getCurrentExtension();
const Convenience = Me.imports.convenience;
const Lib = Me.imports.lib;
let settings = Convenience.getSettings();

function init() {
    if (Main.lookingGlass === null) {
        Main.lookingGlass = new LookingGlass.LookingGlass();
    }
}

/*
 * a recursive function that returns a left-to-right, comma-separated string of
 * children indices that make the path from the ClutterStage root to a given
 * CLutterActor node.
 */
function clutter_path(node) {
    let path = clutter_path_aux(node);
    return path.slice(1);    /* delete leading comma to prepare for later regex split. */
}

function clutter_path_aux(node) {
    let path = '';
    if (node.get_parent() === null) {    /* terminal case is reached at the root. */
        return '';
    } else {
        let n_children = node.get_parent().get_n_children();
        if (n_children === 1) {
            path = clutter_path_aux(node.get_parent()).concat('-0');
        } else {
            for (let i = 0; i < n_children; ++i) {
                if (node === node.get_parent().get_child_at_index(i)) {
                    path = clutter_path_aux(node.get_parent()).concat('-', i);
                    i = n_children;
                }
            }
        }
    }
    return path;
}

/*
 * this function is the inverse of clutter_path(node). null is returned when no
 * actor is found or the 'path' is malformed.
 */
function retrieve_actor(path) {
    if (! /^(\d)+([-](\d)+)*$/.test(path)) {
        return null;    /* format validation*/
    }
    const root = global.stage;
    let actor = root;
    let tuple = path.split('-');
    for (let i = 0; i < tuple.length; ++i) {
        actor = actor.get_child_at_index(tuple[i]);
        if (!actor) {
            return null;
        }
    }
    return actor;
}

const DO_RULES = true;
const UNDO_RULES = false;

/*
 * selectively show or hide all actors on a 'blacklist' depending to its
 * respective intented 'statuses' list and a contextual 'action'.
 * 'action' can be either DO_RULES or UNDO_RULES, and it should be
 * chosen based on the activation state of the whole extension.
 */
function rules(blacklist, statuses, action) {
    if (blacklist != '' && blacklist != '\'\'') {
        let actors = blacklist.split(';');
        let actors_statuses = statuses.split(';');
        let least = ((actors.length < actors_statuses.length) ? actors : actors_statuses);
        let i = 0;
        while (i < least.length) {
            let actor = retrieve_actor(actors[i]);
            if (actor) {
                if (actors_statuses[i] === '1') {
                    if (action === DO_RULES) {
                        actor.hide();
                    } else {
                        actor.show();
                    }
                } else {
                    /* this case is reached when individual widgets' switches are deactivated */
                    if (action === DO_RULES) {
                        actor.show();
                    }
                }
            } else {
                /* log null actor error for later use in GUI */
                Lib.set_actor_error_status(i, '1');
                statuses = settings.get_string('blacklisted-actors-error-statuses');
            }
            ++i;
        }
        /*
         * automatic length mismatch correction, in case settings got corrupted.
         * all the coherent portion of our ruleset has been applied at this point.
         */
        if (actors.length != actors_statuses.length) {
            if (actors_statuses === least) {
                do {
                    statuses.concat(';0');
                    ++i;
                } while (i < actors.length);
                settings.set_string('blacklisted-actors-statuses', statuses);
            } else {
                settings.set_string('blacklisted-actors-statuses', statuses.slice(0, 2*i-1));
            }
        }
    }
}

let ids = {
    mainloop_sources: {
        idle: null,
        timeout: null
    },
    signals: {
        settings: null,
        picker: null
    }
};

var picker = null;

function enable() {
    let blacklist = settings.get_string('blacklisted-actors');
    let statuses = settings.get_string('blacklisted-actors-statuses');
    let error_statuses = settings.get_string('blacklisted-actors-error-statuses');
    ids['mainloop_sources']['idle'] = Mainloop.idle_add(function() {
        ids['mainloop_sources']['timeout'] = Mainloop.timeout_add_seconds(10, function() {
            rules(blacklist, statuses, DO_RULES);
            ids['signals']['settings'] = settings.connect('changed::prefs-modification', function() {
                rules(settings.get_string('blacklisted-actors'), settings.get_string('blacklisted-actors-statuses'), DO_RULES);
            });
            /* ActionMode introduced in 3.16 */
            if (Main.wm.addKeybinding && (Shell.KeyBindingMode || Shell.ActionMode)) {
                let Mode;
                if (Shell.KeyBindingMode === undefined) {
                    Mode = Shell.ActionMode;
                    Mode = Mode.NORMAL | Mode.OVERVIEW | Mode.SYSTEM_MODAL | Mode.POPUP;
                } else {
                    Mode = Shell.KeyBindingMode;
                    Mode = Mode.NORMAL | Mode.OVERVIEW | Mode.MESSAGE_TRAY | Mode.SYSTEM_MODAL | Mode.TOPBAR_POPUP;
                }
                Main.wm.addKeybinding(
                    'actor-picker-shortcut',    /* same key name as in schema, */
                    settings,    /* schema. */
                    Meta.KeyBindingFlags.NONE,
                    Mode,
                    Lang.bind(picker, function() {
                        if (settings.get_boolean('actor-picker-shortcut-status') && picker === null) {
                            picker = new LookingGlass.Inspector(Main.lookingGlass);
                            /*
                             * the callbacks listening for the ESC key signal as specified in
                             * lookingGlass.js do not work well for the peculiar, lookingGlass-less
                             * context we are placing our inspector in. therefore we need to override
                             * them with a global shortcut during the inspector's lifetime and call
                             * the appropiate close() methods from here.
                             */
                            Main.wm.addKeybinding(
                                'actor-picker-close-shortcut',
                                settings,
                                Meta.KeyBindingFlags.NONE,
                                Mode,
                                Lang.bind(picker, function() {
                                    Main.wm.removeKeybinding('actor-picker-close-shortcut');
                                    picker._close();
                                    Main.lookingGlass._objInspector._open = true;
                                    Main.lookingGlass._objInspector.close();
                                    Main.lookingGlass.open();
                                    Main.lookingGlass.close();
                                    picker = null;
                                })
                            );
                            ids['signals']['picker'] = picker.connect('target', Lang.bind(Main.lookingGlass, function(i, target, stageX, stageY) {
                                Main.wm.removeKeybinding('actor-picker-close-shortcut');
                                target.remove_effect(Main.lookingGlass._redBorderEffect);
                                let path = clutter_path(target);
                                /*
                                 * seemingly redundant check, but it will be extremely useful
                                 * to detect bugs and prevent them from getting into the settings.
                                 */
                                if (target === retrieve_actor(path)) {
                                    blacklist = settings.get_string('blacklisted-actors');
                                    statuses = settings.get_string('blacklisted-actors-statuses');
                                    error_statuses = settings.get_string('blacklisted-actors-error-statuses');
                                    settings.set_string(
                                        'blacklisted-actors',
                                        (!(blacklist === '' || blacklist === '\'\'')) ? blacklist.concat(';', path) : path
                                    );
                                    settings.set_string(
                                        'blacklisted-actors-statuses',
                                        (!(statuses === '' || statuses === '\'\'')) ? statuses.concat(';', '1') : '1'
                                    );
                                    settings.set_string(
                                        'blacklisted-actors-error-statuses',
                                        (!(error_statuses === '' || error_statuses === '\'\'')) ? error_statuses.concat(';', '0') : '0'
                                    );
                                    target.hide();
                                    settings.set_boolean('blacklist-addition', !(settings.get_boolean('blacklist-addition')));
                                }
                                picker = null;
                            }));
                        }
                    })
                );
            }
        });
    });
}

function disable() {
    let blacklist = settings.get_string('blacklisted-actors');
    let statuses = settings.get_string('blacklisted-actors-statuses');
    rules(blacklist, statuses, UNDO_RULES);
    Main.wm.removeKeybinding('actor-picker-shortcut');
    settings.disconnect(ids['signals']['settings']);
    /*
     * Explicit Mainloop source removal should be needless when the delayed functions don't
     * return true (i.e. when executed only once), according to API documentation at
     * https://developer.gnome.org/glib/unstable/glib-The-Main-Event-Loop.html
     * and as shown empirically. However,
     * http://blog.mecheye.net/2012/02/requirements-and-tips-for-getting-your-gnome-shell-extension-approved
     * suggests otherwise, so...
     */
    Mainloop.source_remove(ids['mainloop_sources']['timeout']);
    Mainloop.source_remove(ids['mainloop_sources']['idle']);
}
